From c8187969241214a7fde2439e95a8679e17445058 Mon Sep 17 00:00:00 2001
From: Sebastian Pop <spop@amazon.com>
Date: Thu, 29 Jul 2021 17:04:06 +0000
Subject: [PATCH] [arm64] port residual

         residual[4x4]	10.70x 	 7.23     	 77.40
         residual[8x8]	30.92x 	 11.09    	 343.04
       residual[16x16]	48.18x 	 24.47    	 1178.97
       residual[32x32]	56.50x 	 81.18    	 4586.74
---
 source/common/aarch64/asm-primitives.cpp |  6 ++
 source/common/aarch64/pixel-util.S       | 73 ++++++++++++++++++++++++
 2 files changed, 79 insertions(+)

diff --git a/source/common/aarch64/asm-primitives.cpp b/source/common/aarch64/asm-primitives.cpp
index 80558e7..c9b282a 100644
--- a/source/common/aarch64/asm-primitives.cpp
+++ b/source/common/aarch64/asm-primitives.cpp
@@ -413,6 +413,12 @@ void setupAssemblyPrimitives(EncoderPrimitives &p, int cpuMask)
         p.cu[BLOCK_16x16].cpy1Dto2D_shr = PFX(cpy1Dto2D_shr_16x16_neon);
         p.cu[BLOCK_32x32].cpy1Dto2D_shr = PFX(cpy1Dto2D_shr_32x32_neon);
 
+        // calc_Residual
+        p.cu[BLOCK_4x4].calcresidual[NONALIGNED]   = PFX(getResidual4_neon);
+        p.cu[BLOCK_8x8].calcresidual[NONALIGNED]   = PFX(getResidual8_neon);
+        p.cu[BLOCK_16x16].calcresidual[NONALIGNED] = PFX(getResidual16_neon);
+        p.cu[BLOCK_32x32].calcresidual[NONALIGNED] = PFX(getResidual32_neon);
+
         // pixel_var
         p.cu[BLOCK_8x8].var   = PFX(pixel_var_8x8_neon);
         p.cu[BLOCK_16x16].var = PFX(pixel_var_16x16_neon);
diff --git a/source/common/aarch64/pixel-util.S b/source/common/aarch64/pixel-util.S
index 6898e05..c7addf2 100644
--- a/source/common/aarch64/pixel-util.S
+++ b/source/common/aarch64/pixel-util.S
@@ -119,6 +119,79 @@ function x265_pixel_var_64x64_neon
     ret
 endfunc
 
+// void getResidual4_neon(const pixel* fenc, const pixel* pred, int16_t* residual, intptr_t stride)
+function x265_getResidual4_neon
+    lsl             x4, x3, #1
+.rept 2
+    ld1             {v0.8b}, [x0], x3
+    ld1             {v1.8b}, [x1], x3
+    ld1             {v2.8b}, [x0], x3
+    ld1             {v3.8b}, [x1], x3
+    usubl           v4.8h, v0.8b, v1.8b
+    usubl           v5.8h, v2.8b, v3.8b
+    st1             {v4.4h}, [x2], x4
+    st1             {v5.4h}, [x2], x4
+.endr
+    ret
+endfunc
+
+function x265_getResidual8_neon
+    lsl             x4, x3, #1
+.rept 4
+    ld1             {v0.8b}, [x0], x3
+    ld1             {v1.8b}, [x1], x3
+    ld1             {v2.8b}, [x0], x3
+    ld1             {v3.8b}, [x1], x3
+    usubl           v4.8h, v0.8b, v1.8b
+    usubl           v5.8h, v2.8b, v3.8b
+    st1             {v4.8h}, [x2], x4
+    st1             {v5.8h}, [x2], x4
+.endr
+    ret
+endfunc
+
+function x265_getResidual16_neon
+    lsl             x4, x3, #1
+.rept 8
+    ld1             {v0.16b}, [x0], x3
+    ld1             {v1.16b}, [x1], x3
+    ld1             {v2.16b}, [x0], x3
+    ld1             {v3.16b}, [x1], x3
+    usubl           v4.8h, v0.8b, v1.8b
+    usubl2          v5.8h, v0.16b, v1.16b
+    usubl           v6.8h, v2.8b, v3.8b
+    usubl2          v7.8h, v2.16b, v3.16b
+    st1             {v4.8h-v5.8h}, [x2], x4
+    st1             {v6.8h-v7.8h}, [x2], x4
+.endr
+    ret
+endfunc
+
+function x265_getResidual32_neon
+    lsl             x4, x3, #1
+    mov             w12, #4
+.loop_residual_32:
+    sub             w12, w12, #1
+.rept 4
+    ld1             {v0.16b-v1.16b}, [x0], x3
+    ld1             {v2.16b-v3.16b}, [x1], x3
+    ld1             {v4.16b-v5.16b}, [x0], x3
+    ld1             {v6.16b-v7.16b}, [x1], x3
+    usubl           v16.8h, v0.8b, v2.8b
+    usubl2          v17.8h, v0.16b, v2.16b
+    usubl           v18.8h, v1.8b, v3.8b
+    usubl2          v19.8h, v1.16b, v3.16b
+    usubl           v20.8h, v4.8b, v6.8b
+    usubl2          v21.8h, v4.16b, v6.16b
+    usubl           v22.8h, v5.8b, v7.8b
+    usubl2          v23.8h, v5.16b, v7.16b
+    st1             {v16.8h-v19.8h}, [x2], x4
+    st1             {v20.8h-v23.8h}, [x2], x4
+.endr
+    cbnz            w12, .loop_residual_32
+    ret
+endfunc
+
 .macro x265_satd_4x8_8x4_end_neon
     add             v0.8h, v4.8h, v6.8h
     add             v1.8h, v5.8h, v7.8h
-- 
2.25.1

